"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Gateway = void 0;
const cosmwasm_stargate_1 = require("@cosmjs/cosmwasm-stargate");
const connect_sdk_1 = require("@wormhole-foundation/connect-sdk");
const bs58_1 = __importDefault(require("bs58"));
const address_1 = require("./address");
const platform_1 = require("./platform");
var Gateway;
(function (Gateway) {
    Gateway.name = "Wormchain";
    function gatewayAddress() {
        const contracts = platform_1.CosmwasmPlatform.contracts.getContracts(Gateway.name);
        return contracts.gateway;
    }
    Gateway.gatewayAddress = gatewayAddress;
    function tokenBridgeAddress() {
        const contracts = platform_1.CosmwasmPlatform.contracts.getContracts(Gateway.name);
        return contracts.tokenBridge;
    }
    Gateway.tokenBridgeAddress = tokenBridgeAddress;
    function coreAddress() {
        const contracts = platform_1.CosmwasmPlatform.contracts.getContracts(Gateway.name);
        return contracts.coreBridge;
    }
    Gateway.coreAddress = coreAddress;
    // Returns RPC client for Wormchain
    async function getRpc() {
        const rpcAddress = platform_1.CosmwasmPlatform.conf[Gateway.name].rpc;
        return await cosmwasm_stargate_1.CosmWasmClient.connect(rpcAddress);
    }
    Gateway.getRpc = getRpc;
    async function getTokenBridge(rpc) {
        rpc = rpc ? rpc : await getRpc();
        return platform_1.CosmwasmPlatform.getTokenBridge(rpc);
    }
    Gateway.getTokenBridge = getTokenBridge;
    async function getIbcBridge(rpc) {
        rpc = rpc ? rpc : await getRpc();
        return platform_1.CosmwasmPlatform.getIbcBridge(rpc);
    }
    Gateway.getIbcBridge = getIbcBridge;
    // Get the wrapped version of an asset created on wormchain
    // for a given chain
    async function getWrappedAsset(token) {
        const tb = await getTokenBridge();
        const wrappedAsset = await tb.getWrappedAsset(token);
        // Encode the original address to base58 and add it
        // to the factory address for cw20 style factory token address
        const encodedAddress = bs58_1.default.encode(wrappedAsset.toUniversalAddress().toUint8Array());
        const factoryAddress = `factory/${gatewayAddress()}/${encodedAddress}`;
        return new address_1.CosmwasmAddress(factoryAddress);
    }
    Gateway.getWrappedAsset = getWrappedAsset;
    // Returns the destination channel on wormchain for given source chain
    function getDestinationChannel(chain) {
        const channels = platform_1.CosmwasmPlatform.getIbcChannel(chain);
        if (!channels)
            throw new Error("No channels configured for chain " + chain);
        return channels.dstChannel;
    }
    Gateway.getDestinationChannel = getDestinationChannel;
    // Gets the source channel on wormchain for a given chain
    function getSourceChannel(chain) {
        const channels = platform_1.CosmwasmPlatform.getIbcChannel(chain);
        if (!channels)
            throw new Error("No channels configured for chain " + chain);
        return channels.srcChannel;
    }
    Gateway.getSourceChannel = getSourceChannel;
    // Returns whether or not a given chain is gateway supported
    function isSupported(chain) {
        return platform_1.CosmwasmPlatform.getIbcChannel(chain) !== null;
    }
    Gateway.isSupported = isSupported;
    // Derive the Token Address with context for whether or not its managed
    // https://github.com/wormhole-foundation/wormhole/blob/251e6c4a6478379ff862aed08d835f9022ef4143/cosmwasm/contracts/token-bridge/src/token_address.rs#L12
    function deriveTokenAddress(chain, asset) {
        const tokenId = new Uint8Array(32);
        //const addr = toNative(chain, asset) as CosmwasmAddress;
        const nativeFlg = platform_1.CosmwasmPlatform.isNativeDenom(chain, asset) ? 1 : 0;
        tokenId.set([nativeFlg], 0);
        tokenId.set((0, connect_sdk_1.keccak256)(asset).slice(1), 1);
        return tokenId;
    }
    Gateway.deriveTokenAddress = deriveTokenAddress;
    // derive the ics20 token denom from the
    // wrapped denom and destination channel
    function deriveIbcDenom(chain, denom) {
        const channel = getDestinationChannel(chain);
        const hashData = Buffer.from(`transfer/${channel}/${denom}`);
        const hash = Buffer.from((0, connect_sdk_1.sha256)(hashData)).toString("hex");
        return new address_1.CosmwasmAddress(`ibc/${hash.toUpperCase()}`);
    }
    Gateway.deriveIbcDenom = deriveIbcDenom;
    function makeTransferMsg(chain, recipient, fee = 0n, payload, nonce) {
        // Address of recipient is b64 encoded Cosmos bech32 address
        const address = Buffer.from(recipient.toString()).toString("base64");
        const common = {
            chain: (0, connect_sdk_1.toChainId)(chain),
            recipient: address,
            fee: fee.toString(),
            nonce: nonce ?? Math.round(Math.random() * 100000),
        };
        const msg = payload
            ? {
                gateway_transfer_with_payload: { ...common, payload: payload },
            }
            : { gateway_transfer: { ...common } };
        return msg;
    }
    Gateway.makeTransferMsg = makeTransferMsg;
    // TODO: make consts
    function getWormholeMessage(tx) {
        const wasm = tx.events
            .filter((ev) => ev.type === "wasm" &&
            ev.attributes[0].key === "_contract_address" &&
            ev.attributes[0].value === coreAddress())
            .pop();
        const sequence = wasm.attributes.find((e) => {
            return e.key === "message.sequence";
        }).value;
        const emitter = wasm.attributes.find((e) => {
            return e.key === "message.sender";
        }).value;
        return {
            chain: Gateway.name,
            emitter: new connect_sdk_1.UniversalAddress(emitter),
            sequence: BigInt(sequence),
        };
    }
    Gateway.getWormholeMessage = getWormholeMessage;
})(Gateway || (exports.Gateway = Gateway = {}));
//# sourceMappingURL=gateway.js.map