"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CosmwasmIbcBridge = void 0;
const stargate_1 = require("@cosmjs/stargate");
const connect_sdk_1 = require("@wormhole-foundation/connect-sdk");
const tx_1 = require("cosmjs-types/ibc/applications/transfer/v1/tx");
const constants_1 = require("../constants");
const gateway_1 = require("../gateway");
const platform_1 = require("../platform");
const platformUtils_1 = require("../platformUtils");
const unsignedTransaction_1 = require("../unsignedTransaction");
const millisToNano = (seconds) => seconds * 1000000;
class CosmwasmIbcBridge {
    network;
    chain;
    rpc;
    contracts;
    gatewayAddress;
    gatewayChannel;
    // map the local channel ids to the remote chain
    channelMap = new Map();
    constructor(network, chain, rpc, contracts) {
        this.network = network;
        this.chain = chain;
        this.rpc = rpc;
        this.contracts = contracts;
        this.gatewayAddress = this.contracts.getContracts(gateway_1.Gateway.name).gateway;
        const isGateway = this.chain === gateway_1.Gateway.name;
        for (const [chain, channel] of (0, constants_1.networkToChannelMap)(network)) {
            if (!isGateway && this.chain === chain)
                this.gatewayChannel = channel.dstChannel;
            this.channelMap.set(channel[isGateway ? "dstChannel" : "srcChannel"], chain);
        }
    }
    static async fromProvider(rpc, contracts) {
        const [network, chain] = await platform_1.CosmwasmPlatform.chainFromRpc(rpc);
        return new CosmwasmIbcBridge(network, chain, rpc, contracts);
    }
    async *transfer(sender, recipient, token, amount) {
        const senderAddress = sender.toString();
        const nonce = Math.round(Math.random() * 10000);
        // TODO: needs heavy testing
        let recipientAddress;
        if ((0, connect_sdk_1.chainToPlatform)(recipient.chain) === "Cosmwasm") {
            // If cosmwasm, we just want the b64 encoded string address
            recipientAddress = Buffer.from(recipient.address.toString()).toString("base64");
        }
        else {
            // If we're bridging out of cosmos, we need the universal address
            recipientAddress = Buffer.from(recipient.address.toUniversalAddress().toUint8Array()).toString("base64");
        }
        const payload = {
            gateway_ibc_token_bridge_payload: {
                gateway_transfer: {
                    recipient: recipientAddress,
                    chain: (0, connect_sdk_1.toChainId)(recipient.chain),
                    nonce,
                    // TODO: fetch param of which contract?
                    fee: "0",
                },
            },
        };
        const timeout = millisToNano(Date.now() + constants_1.IBC_TIMEOUT_MILLIS);
        const memo = JSON.stringify(payload);
        const ibcDenom = gateway_1.Gateway.deriveIbcDenom(this.chain, token.toString());
        const ibcToken = (0, stargate_1.coin)(amount.toString(), ibcDenom.toString());
        const ibcMessage = {
            typeUrl: constants_1.IBC_MSG_TYPE,
            value: tx_1.MsgTransfer.fromPartial({
                sourcePort: constants_1.IBC_TRANSFER_PORT,
                sourceChannel: this.gatewayChannel,
                sender: senderAddress,
                receiver: this.gatewayAddress,
                token: ibcToken,
                timeoutTimestamp: timeout,
                memo,
            }),
        };
        yield this.createUnsignedTx({
            msgs: [ibcMessage],
            fee: (0, unsignedTransaction_1.computeFee)(this.chain),
            memo: "Wormhole.TransferToGateway",
        }, "IBC.transfer");
        return;
    }
    async lookupTransferFromTx(txid) {
        const txResults = await this.rpc.getTx(txid);
        if (!txResults)
            throw new Error(`No transaction found with txid: ${txid}`);
        if (txResults.code !== 0)
            throw new Error(`Transaction failed: ${txResults.rawLog}`);
        const xfers = await this.fetchTransferInfo(txResults);
        if (xfers.length === 0)
            throw new Error("No transfers found for tx: " + txid);
        if (xfers.length > 1)
            console.error(">1 xfer in tx; why");
        return xfers[0];
    }
    async lookupMessageFromIbcMsgId(msg) {
        const tx = await this.lookupTxFromIbcMsgId(msg);
        return gateway_1.Gateway.getWormholeMessage(tx);
    }
    async lookupTxFromIbcMsgId(msg) {
        const prefix = this.chain === msg.chain ? constants_1.IBC_PACKET_SEND : constants_1.IBC_PACKET_RECEIVE;
        const { srcChannel, dstChannel, sequence, srcPort, dstPort } = msg;
        // Find the transaction with matching payload
        const txResults = await this.rpc.searchTx([
            {
                key: `${prefix}.${constants_1.IBC_PACKET_DST}`,
                value: dstChannel,
            },
            {
                key: `${prefix}.${constants_1.IBC_PACKET_SRC}`,
                value: srcChannel,
            },
            {
                key: `${prefix}.${constants_1.IBC_PACKET_SRC_PORT}`,
                value: srcPort,
            },
            {
                key: `${prefix}.${constants_1.IBC_PACKET_DST_PORT}`,
                value: dstPort,
            },
            {
                key: `${prefix}.${constants_1.IBC_PACKET_SEQ}`,
                value: sequence.toString(),
            },
        ]);
        if (txResults.length === 0)
            throw new Error(`Found no transactions for message: ` + JSON.stringify(msg));
        if (txResults.length > 1)
            console.error(`Expected 1 transaction, got ${txResults.length} found for IbcMsgid: ${msg}`);
        const [tx] = txResults;
        return tx;
    }
    async lookupTransferFromIbcMsgId(msg) {
        // Finds the transaction but there may be multiple
        // IBCTransfers as part of this
        const tx = await this.lookupTxFromIbcMsgId(msg);
        const xfers = await this.fetchTransferInfo(tx);
        if (xfers.length === 0)
            throw new Error(`No transfers found on ${this.chain} in tx: ${tx}`);
        // TODO
        if (xfers.length > 1)
            console.error(">1 xfer in tx; why");
        return xfers[0];
    }
    // Returns the IBC Transfer message content and IBC transfer information
    async lookupTransferFromMsg(msg) {
        const encodedPayload = Buffer.from(JSON.stringify(msg)).toString("base64");
        // Find the transaction with matching payload
        const txResults = await this.rpc.searchTx([
            {
                key: "wasm.transfer_payload",
                value: encodedPayload,
            },
        ]);
        if (txResults.length === 0)
            throw new Error(`Found no transactions for payload: ` + JSON.stringify(encodedPayload));
        if (txResults.length !== 1)
            console.error("Expected 1 tx, got: ", txResults.length);
        const [tx] = txResults;
        const xfers = await this.fetchTransferInfo(tx);
        if (xfers.length === 0)
            throw new Error(`Found no transactions for payload: ` + JSON.stringify(encodedPayload));
        if (xfers.length !== 1)
            console.error("Expected 1 xfer, got: ", xfers.length);
        return xfers[0];
    }
    parseIbcTransferInfo(tx) {
        // Try to get all IBC packets (sent/received)
        const packets = tx.events.filter((ev) => ev.type === constants_1.IBC_PACKET_SEND || ev.type === constants_1.IBC_PACKET_RECEIVE);
        if (packets.length === 0)
            throw new Error(`No IBC Transfers on ${this.chain} found in: ${tx.hash}`);
        // TODO: gotta be a way to parse these out better?
        // Try to assemble attributes from packet fields
        const xfers = new Set();
        for (const packet of packets) {
            const xfer = { pending: false };
            const msgId = {};
            for (const attr of packet.attributes) {
                // set on msgId
                if (attr.key === constants_1.IBC_PACKET_SRC)
                    msgId.srcChannel = attr.value;
                if (attr.key === constants_1.IBC_PACKET_DST)
                    msgId.dstChannel = attr.value;
                if (attr.key === constants_1.IBC_PACKET_SEQ)
                    msgId.sequence = Number(attr.value);
                if (attr.key === constants_1.IBC_PACKET_SRC_PORT)
                    msgId.srcPort = attr.value;
                if (attr.key === constants_1.IBC_PACKET_DST_PORT)
                    msgId.dstPort = attr.value;
                // set on the xfer obj
                if (attr.key === constants_1.IBC_PACKET_DATA)
                    xfer.data = JSON.parse(attr.value);
            }
            // If we're receiving a packet, we need figure out who sent it
            // possibly resolving by source channel
            // we assign the chain to the sender as iternally canonical
            msgId.chain =
                packet.type === constants_1.IBC_PACKET_SEND
                    ? this.chain
                    : this.channelMap.get(msgId.srcChannel);
            // Note: using the type guard to tell us if we have all the fields we expect
            if ((0, connect_sdk_1.isIbcMessageId)(msgId))
                xfer.id = msgId;
            else
                throw new Error("Invalid IbcMessageId: " + JSON.stringify(msgId));
            if ((0, connect_sdk_1.isIbcTransferInfo)(xfer))
                xfers.add(xfer);
            else
                throw new Error("Invalid IbcTransferInfo: " + JSON.stringify(xfer));
        }
        return Array.from(xfers);
    }
    // fetch whether or not this transfer is pending
    async fetchTransferInfo(tx) {
        // Try to get all IBC packets (sent/received)
        const xfers = this.parseIbcTransferInfo(tx);
        const transfers = [];
        for (const xfer of xfers) {
            // If its present in the commitment results, its interpreted as in-flight
            // the client throws an error and we report any error as not in-flight
            const qc = platformUtils_1.CosmwasmUtils.asQueryClient(this.rpc);
            try {
                await qc.ibc.channel.packetCommitment(constants_1.IBC_TRANSFER_PORT, xfer.id.srcChannel, xfer.id.sequence);
                xfer.pending = true;
            }
            catch (e) {
                // TODO: catch http errors unrelated to no commitment in flight
                // otherwise we might lie about pending = false
            }
            // TODO other states of packet in-flight-ness?
            transfers.push(xfer);
        }
        return transfers;
    }
    // Fetches the channel information between wormchain and a given chain
    async fetchChannel(chain) {
        const queryClient = platformUtils_1.CosmwasmUtils.asQueryClient(this.rpc);
        try {
            const { channel: srcChannel } = await this.rpc.queryContractSmart(this.gatewayAddress, { ibc_channel: { chain_id: (0, connect_sdk_1.toChainId)(chain) } });
            const conn = await queryClient.ibc.channel.channel(constants_1.IBC_TRANSFER_PORT, srcChannel);
            const dstChannel = conn.channel?.counterparty?.channelId;
            if (!dstChannel)
                throw new Error(`No destination channel found for chain ${chain} on ${this.chain}`);
            return { srcChannel, dstChannel };
        }
        catch (e) {
            // TODO
        }
        return null;
    }
    createUnsignedTx(txReq, description, parallelizable = false) {
        return new unsignedTransaction_1.CosmwasmUnsignedTransaction(txReq, this.network, this.chain, description, parallelizable);
    }
}
exports.CosmwasmIbcBridge = CosmwasmIbcBridge;
//# sourceMappingURL=ibc.js.map