"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CosmwasmTokenBridge = void 0;
const connect_sdk_1 = require("@wormhole-foundation/connect-sdk");
const unsignedTransaction_1 = require("../unsignedTransaction");
const platform_1 = require("../platform");
class CosmwasmTokenBridge {
    network;
    chain;
    rpc;
    contracts;
    tokenBridge;
    translator;
    constructor(network, chain, rpc, contracts) {
        this.network = network;
        this.chain = chain;
        this.rpc = rpc;
        this.contracts = contracts;
        this.tokenBridge = this.contracts.getTokenBridge(this.chain, this.rpc);
        if (connect_sdk_1.contracts.translator.has(network, chain)) {
            this.translator = connect_sdk_1.contracts.translator.get(network, chain);
        }
    }
    static async fromProvider(rpc, contracts) {
        const [network, chain] = await platform_1.CosmwasmPlatform.chainFromRpc(rpc);
        return new CosmwasmTokenBridge(network, chain, rpc, contracts);
    }
    async isWrappedAsset(token) {
        try {
            await this.getOriginalAsset(token);
            return true;
        }
        catch { }
        return false;
    }
    async hasWrappedAsset(token) {
        try {
            await this.getWrappedAsset(token);
            return true;
        }
        catch { }
        return false;
    }
    async getWrappedAsset(token) {
        if (token.chain === this.chain)
            throw new Error(`Expected foreign chain, got ${token.chain}`);
        const base64Addr = Buffer.from(token.address.toUniversalAddress().toUint8Array()).toString("base64");
        const { address } = await this.rpc.queryContractSmart(this.tokenBridge, {
            wrapped_registry: {
                chain: (0, connect_sdk_1.toChainId)(token.chain),
                address: base64Addr,
            },
        });
        // @ts-ignore
        return (0, connect_sdk_1.toNative)(this.chain, address);
    }
    async getOriginalAsset(token) {
        const wrappedAddress = token.toString();
        const response = await this.rpc.queryContractSmart(wrappedAddress, {
            wrapped_asset_info: {},
        });
        const origChain = (0, connect_sdk_1.toChainName)(response.asset_chain);
        const origAddress = Buffer.from(response.asset_address, "base64");
        return {
            chain: origChain,
            address: new connect_sdk_1.UniversalAddress(new Uint8Array(origAddress)),
        };
    }
    async isTransferCompleted(vaa) {
        const data = Buffer.from((0, connect_sdk_1.serialize)(vaa)).toString("base64");
        const result = await this.rpc.queryContractSmart(this.tokenBridge, {
            is_vaa_redeemed: { vaa: data },
        });
        return result.is_redeemed;
    }
    async *createAttestation(token, payer) {
        if (!payer)
            throw new Error("Payer required to create attestation");
        // TODO nonce?
        const nonce = 0;
        const assetInfo = token === "native"
            ? {
                native_token: {
                    denom: platform_1.CosmwasmPlatform.getNativeDenom(this.chain),
                },
            }
            : {
                token: { contract_addr: token.toString() },
            };
        yield this.createUnsignedTx({
            msgs: [
                (0, unsignedTransaction_1.buildExecuteMsg)(payer.toString(), this.tokenBridge, {
                    create_asset_meta: { asset_info: assetInfo, nonce },
                }),
            ],
            fee: (0, unsignedTransaction_1.computeFee)(this.chain),
            memo: "Wormhole - Create Attestation",
        }, "TokenBridge.createAttestation");
    }
    async *submitAttestation(vaa, payer) {
        if (!payer)
            throw new Error("Payer required to submit attestation");
        yield this.createUnsignedTx({
            msgs: [
                (0, unsignedTransaction_1.buildExecuteMsg)(payer.toString(), this.tokenBridge, {
                    submit_vaa: { data: (0, connect_sdk_1.serialize)(vaa) },
                }),
            ],
            fee: (0, unsignedTransaction_1.computeFee)(this.chain),
            memo: "Wormhole - Submit Attestation",
        }, "TokenBridge.submitAttestation");
    }
    async *transfer(sender, recipient, token, amount, payload) {
        const nonce = Math.round(Math.random() * 100000);
        const relayerFee = "0";
        const recipientChainId = (0, connect_sdk_1.toChainId)(recipient.chain);
        const recipientAddress = Buffer.from(recipient.address.toUniversalAddress().toUint8Array());
        const denom = platform_1.CosmwasmPlatform.getNativeDenom(this.chain);
        const isNative = token === "native";
        const tokenAddress = isNative ? denom : token.toString();
        const senderAddress = sender.toString();
        const mk_initiate_transfer = (info) => {
            const common = {
                asset: {
                    amount: amount.toString(),
                    info,
                },
                recipient_chain: recipientChainId,
                recipient: Buffer.from(recipientAddress).toString("base64"),
                fee: relayerFee,
                nonce: nonce,
            };
            return payload
                ? {
                    initiate_transfer_with_payload: { ...common, payload },
                }
                : {
                    initiate_transfer: common,
                };
        };
        if (isNative) {
            const msgs = [
                (0, unsignedTransaction_1.buildExecuteMsg)(senderAddress, this.tokenBridge, { deposit_tokens: {} }, [{ amount: amount.toString(), denom: tokenAddress }]),
                (0, unsignedTransaction_1.buildExecuteMsg)(senderAddress, this.tokenBridge, mk_initiate_transfer({
                    native_token: { denom: tokenAddress },
                })),
            ];
            yield this.createUnsignedTx({
                msgs,
                fee: (0, unsignedTransaction_1.computeFee)(this.chain),
                memo: "Wormhole - Initiate Native Transfer",
            }, "TokenBridge.transferNative");
        }
        else {
            const msgs = [
                (0, unsignedTransaction_1.buildExecuteMsg)(senderAddress, tokenAddress, {
                    increase_allowance: {
                        spender: this.tokenBridge,
                        amount: amount.toString(),
                        expires: { never: {} },
                    },
                }),
                (0, unsignedTransaction_1.buildExecuteMsg)(senderAddress, this.tokenBridge, mk_initiate_transfer({
                    token: { contract_addr: tokenAddress },
                })),
            ];
            yield this.createUnsignedTx({
                msgs,
                fee: (0, unsignedTransaction_1.computeFee)(this.chain),
                memo: "Wormhole - Initiate Transfer",
            }, "TokenBridge.transfer");
        }
        return;
    }
    async *redeem(sender, vaa, unwrapNative = true) {
        // TODO: unwrapNative
        const data = Buffer.from((0, connect_sdk_1.serialize)(vaa)).toString("base64");
        let msg;
        if (this.translator &&
            (0, connect_sdk_1.toNative)(this.chain, this.translator)
                .toUniversalAddress()
                .equals(vaa.payload.to.address)) {
            msg = (0, unsignedTransaction_1.buildExecuteMsg)(sender.toString(), this.translator, {
                complete_transfer_and_convert: {
                    vaa: data,
                },
            });
        }
        else {
            msg = (0, unsignedTransaction_1.buildExecuteMsg)(sender.toString(), this.tokenBridge, {
                submit_vaa: { data },
            });
        }
        yield this.createUnsignedTx({
            msgs: [msg],
            fee: (0, unsignedTransaction_1.computeFee)(this.chain),
            memo: "Wormhole - Complete Transfer",
        }, "TokenBridge.redeem");
        return;
    }
    async parseTransactionDetails(txid) {
        throw new Error("Not implemented");
    }
    async getWrappedNative() {
        return (0, connect_sdk_1.toNative)(this.chain, platform_1.CosmwasmPlatform.getNativeDenom(this.chain));
    }
    createUnsignedTx(txReq, description, parallelizable = false) {
        return new unsignedTransaction_1.CosmwasmUnsignedTransaction(txReq, this.network, this.chain, description, parallelizable);
    }
}
exports.CosmwasmTokenBridge = CosmwasmTokenBridge;
//# sourceMappingURL=tokenBridge.js.map